/* ts7200.c
 *
 * $Id:$
 *
 * Physical device output support for TS7200-series ARM boards.
 *
 * Copyright (C) by Matt Godbolt (themoog@xania.org)
 * Based on Embedded ARM's sample LCD driver code.
 * Released under the terms of the GNU GPL, see file COPYING for more details.
 */

#include "ts7200.h"

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/ioport.h>
#include <linux/errno.h>
#include <linux/delay.h>
#include <asm/io.h>

#ifdef PCMASK
#undef PCMASK
#endif

#define GPIOBASE	0x80840000
#define GPIOSIZE	0x000000cc
#define MODELBASE   0x22000000
#define MODELSIZE   0x00000004
#define TS7300		0x03
#define PADR		0x00
#define PADDR		0x10
#define PAMASK		0x7F
#define PCDR		0x08
#define PCDDR		0x18
#define PCMASK		0x01
#define PHDR		0x40
#define PHDDR		0x44

#define READ(OFFSET) readl((volatile uint32_t*)(gpio_ptr + OFFSET))
#define WRITE(OFFSET, DATA) writel(DATA, (volatile uint32_t*)(gpio_ptr + OFFSET))

#define SETUP	300
#define PULSE	800
#define HOLD	400

static volatile char *gpio_ptr;
static uint32_t model;

int ts7200_initialise(void)
{
	volatile char *model_base;

	/* NB this is untested for 2.6 kernels; cribbed from gps_pps code supplied on 7200 mailing list. */
	gpio_ptr = (volatile char *)__ioremap(GPIOBASE, GPIOSIZE, 0);
	if (!gpio_ptr)
	{
		printk(KERN_ERR "LCD: Unable to map GPIO pointer\n");
		return 0;
	}
	printk(KERN_DEBUG "LCD: gpio_base at 0x%08x\n", (int)gpio_ptr);

	model_base = (volatile char *)__ioremap(MODELBASE, MODELSIZE, 0);
	if (!model_base)
	{
		iounmap((void *)gpio_ptr);
		gpio_ptr = 0;
		printk(KERN_ERR "LCD: Unable to map MODELBASE pointer\n");
		return 0;
	}
	model = readl(model_base) & 0x03;
	iounmap((void *)model_base);
	printk(KERN_DEBUG "LCD: found ts7200 board model %d\n", model);

	/* bits 3:5 of port H to outputs */
	WRITE(PHDDR, READ(PHDDR) | 0x38);
	/* de-assert EN, de-assert RS */
	WRITE(PHDR, READ(PHDR) &~ 0x18);
	udelay(15000);
	ts7200_command(0x38); /* two rows, 5x7, 8 bit */
	udelay(4100);
	ts7200_command(0x38); /* two rows, 5x7, 8 bit */
	udelay(100);
	ts7200_command(0x38); /* two rows, 5x7, 8 bit */
	return 1;
}

void ts7200_finalise()
{
	if (gpio_ptr)
		iounmap((void *)gpio_ptr);
}

static void ts7200_command_or_data(int command, int is_command)
{
	uint32_t control = READ(PHDR);

	if (model == TS7300)
	{
		WRITE(PADDR, READ(PADDR) | PAMASK); /* set port A to outputs */
		WRITE(PADDR, READ(PCDDR) | PCMASK); /* set port C to outputs */
	}
	else
	{
		WRITE(PADDR, 0xff); /* set port A to outputs */
	}

	/* step 1, apply RS & WR, send data */
	if (model == TS7300)
	{
		WRITE(PADR, (READ(PADR) &~PAMASK) | (command & PAMASK));
		WRITE(PCDR, (READ(PCDR) &~PCMASK) | (command >> 7));
	}
	else
	{
		WRITE(PADR, command);
	}

	if (is_command)
	{
		control &= ~0x30; /* de-assert RS, assert WR */
	}
	else
	{
		control |= 0x10; /* assert RS */
		control &= ~0x20; /* assert WR */
	}
	WRITE(PHDR, control);

	/* step 2, wait */
	ndelay(SETUP);

	/* step 3, assert EN */
	control |= 0x8;
	WRITE(PHDR, control);

	/* step 4, wait */
	ndelay(PULSE);

	/* step 5, de-assert EN	*/
	control &= ~0x8;
	WRITE(PHDR, control);

	/* step 6, wait */
	ndelay(HOLD);
}

void ts7200_wait(void)
{
	int dat, tries = 0;
	uint32_t control;

	if (model == TS7300)
	{
		WRITE(PADDR, 0); /* port A to inputs */
		WRITE(PCDDR, 0); /* port C to inputs */
	}
	else
	{
		WRITE(PADDR, 0); /* port A to inputs */
	}

	control = READ(PHDR);

	do
	{
		/* step 1, apply RS & WR */
		control |= 0x30; /* de-assert WR */
		control &= ~0x10; /* de-assert RS */
		WRITE(PHDR, control);

		/* step 2, wait */
		ndelay(SETUP);

		/* step 3, assert EN */
		control |= 0x8;
		WRITE(PHDR, control);

		/* step 4, wait */
		ndelay(PULSE);

		/* step 5, de-assert EN, read result */
		if (model == TS7300)
		{
			dat = (READ(PADR) & PAMASK) | ((READ(PCDR) & PCMASK) << 7);
		}
		else
		{
			dat = READ(PADR);
		}

		/* de-assert EN */
		control &= ~0x8;
		WRITE(PHDR, control);

		// step 6, wait
		ndelay(HOLD);
	} while (dat & 0x80 && tries++ < 1000);
}

void ts7200_command(int command)
{
	ts7200_command_or_data(command, 1);
}

void ts7200_data(int data)
{
	ts7200_wait();
	ts7200_command_or_data(data, 0);
}
